# Generated by Django 5.0.13 on 2025-03-20 08:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbportfolio', '0074_alter_rebalancer_frequency_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='portfolio',
            name='initial_position_date',
            field=models.DateField(blank=True, null=True, verbose_name='Last Position Date'),
        ),
        migrations.AddField(
            model_name='portfolio',
            name='last_position_date',
            field=models.DateField(blank=True, null=True, verbose_name='Last Position Date'),
        ),
        migrations.AlterField(
            model_name='trade',
            name='transaction_subtype',
            field=models.CharField(choices=[('REBALANCE', 'Rebalance'), ('DECREASE', 'Decrease'), ('INCREASE', 'Increase'), ('SUBSCRIPTION', 'Subscription'), ('REDEMPTION', 'Redemption'), ('BUY', 'Buy'), ('SELL', 'Sell'), ('NO_CHANGE', 'No Change')], default='BUY', max_length=32, verbose_name='Trade Type'),
        ),
    ]
