# Generated by Django 5.0.14 on 2025-08-04 09:30

from decimal import Decimal
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbportfolio', '0085_order_desired_target_weight'),
    ]

    operations = [
        migrations.AddField(
            model_name='orderproposal',
            name='total_cash_weight',
            field=models.DecimalField(decimal_places=4, default=Decimal('0'), help_text='The desired percentage for the cash component. The remaining percentage (100% minus this value) will be allocated to total target weighting. Default is 0%.', max_digits=5, verbose_name='Total Cash Weight'),
        ),
    ]
