# Generated by Django 4.2.11 on 2024-03-14 15:54

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbfdm", "0013_instrument_is_investable_universe_and_more"),
        ("wbportfolio", "0056_remove_companyportfoliodata_assets_under_management_currency_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="portfolio",
            name="preferred_instrument_classifications",
            field=models.ManyToManyField(
                blank=True,
                limit_choices_to=models.Q(("instrument_type__is_classifiable", True), ("level", 0)),
                related_name="preferred_portfolio_classifications",
                through="wbportfolio.PortfolioInstrumentPreferredClassificationThroughModel",
                to="wbfdm.instrument",
                verbose_name="The Preferred classification per instrument",
            ),
        ),
        migrations.AlterField(
            model_name="portfolioinstrumentpreferredclassificationthroughmodel",
            name="instrument",
            field=models.ForeignKey(
                limit_choices_to=models.Q(("instrument_type__is_classifiable", True), ("level", 0)),
                on_delete=django.db.models.deletion.CASCADE,
                related_name="preferred_classification_portfolio",
                to="wbfdm.instrument",
            ),
        ),
    ]
