# Generated by Django 5.0.14 on 2025-08-18 08:34

from decimal import Decimal
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbportfolio', '0087_product_order_routing_custodian_adapter'),
    ]

    operations = [
        migrations.AddField(
            model_name='orderproposal',
            name='total_effective_portfolio_contribution',
            field=models.DecimalField(decimal_places=16, default=Decimal('1'), max_digits=19),
        ),
    ]
